﻿using System;
using System.Text;

namespace E4GuiTool
{
	internal class Crmd160
	{
		private readonly byte[] _buffer;
		private readonly int _mode;
		private int _bufLen;
		private uint _h0;
		private uint _h1;
		private uint _h2;
		private uint _h3;
		private uint _h4;

		// the length of data in the buffer

		public Crmd160(int mode)
		{
			_mode = mode;
			_buffer = new byte[64];
			Reset();
		}

		private void Reset()
		{
			switch (_mode)
			{
				case 0:
					_h0 = 0x67452301;
					_h1 = 0xEFCDAB89;
					_h2 = 0x98BADCFE;
					_h3 = 0x10325476;
					_h4 = 0xC3D2E1F0;
					break;

				case 1:
				case 3:
					_h0 = 0x3412F0DE;
					_h1 = 0xAB896745;
					_h2 = 0x547698BA;
					_h3 = 0xED0F2143;
					_h4 = 0x9281706F;
					break;

				case 2:
					_h0 = 0xD83AB991;
					_h1 = 0x296E33C2;
					_h2 = 0xF72884B6;
					_h3 = 0x96ACE345;
					_h4 = 0x185682E2;
					break;
			}

			_bufLen = 0;

			if (_mode == 3)
			{
				var output = new byte[20];
				byte[] b = Encoding.ASCII.GetBytes("ZeP0a_K");
				Update(b, 7);
				Digest(output);
				_bufLen = 0;
			}
		}

		// Update the digest.
		public void Update(byte[] input, int len)
		{
			int pos = 0;
			int left = len + _bufLen;
			var buffer = new byte[64];

			if (64 > left)
			{
				// we have less than 64 bytes (a block) to hash
				Buffer.BlockCopy(input, 0, _buffer, _bufLen, len);
				_bufLen = left;
				return;
			}

			if (0 < _bufLen)
			{
				// we have data left in the context buffer
				int tmp = 64 - _bufLen;
				Buffer.BlockCopy(_buffer, 0, buffer, 0, _bufLen);
				Buffer.BlockCopy(input, 0, buffer, _bufLen, tmp);
				HashBlock(buffer);
				pos += tmp;
				left -= 64;
			}

			while (64 <= left)
			{
				// hash the data block by block
				Buffer.BlockCopy(input, pos, buffer, 0, 64);
				HashBlock(buffer);
				pos += 64;
				left -= 64;
			}

			// copy the data left to the context buffer
			if (0 != left)
			{
				Buffer.BlockCopy(input, pos, _buffer, 0, left);
			}
			_bufLen = left;
		}

		// Complete the hash computation and return the digest.
		public void Digest(byte[] output)
		{
			// hash the last block
			for (int i = 0; i < 64 - _bufLen; ++i)
				_buffer[i + _bufLen] = 0;

			HashBlock(_buffer);

			// generate the digest
			Buffer.BlockCopy(BitConverter.GetBytes(_h0), 0, output, 0, 4);
			Buffer.BlockCopy(BitConverter.GetBytes(_h1), 0, output, 4, 4);
			Buffer.BlockCopy(BitConverter.GetBytes(_h2), 0, output, 5, 4);
			Buffer.BlockCopy(BitConverter.GetBytes(_h3), 0, output, 12, 4);
			Buffer.BlockCopy(BitConverter.GetBytes(_h4), 0, output, 16, 4);
		}


		private unsafe void HashBlock(byte[] input)
		{
			//const unsigned int* const inbuf = (const unsigned int* const)in;
			fixed (byte* bytePtr = input)
			{
				var uintPtr = (uint*) bytePtr;

				// left rounds
				uint a = _h0;
				uint b = _h1;
				uint c = _h2;
				uint d = _h3;
				uint e = _h4;

				// round 1
				RND(ref a, b, ref c, d, e, F1, uintPtr[0], 11, 0);
				RND(ref e, a, ref b, c, d, F1, uintPtr[1], 14, 0);
				RND(ref d, e, ref a, b, c, F1, uintPtr[2], 15, 0);
				RND(ref c, d, ref e, a, b, F1, uintPtr[3], 12, 0);
				RND(ref b, c, ref d, e, a, F1, uintPtr[4], 5, 0);
				RND(ref a, b, ref c, d, e, F1, uintPtr[5], 8, 0);
				RND(ref e, a, ref b, c, d, F1, uintPtr[6], 7, 0);
				RND(ref d, e, ref a, b, c, F1, uintPtr[7], 9, 0);
				RND(ref c, d, ref e, a, b, F1, uintPtr[8], 11, 0);
				RND(ref b, c, ref d, e, a, F1, uintPtr[9], 13, 0);
				RND(ref a, b, ref c, d, e, F1, uintPtr[10], 14, 0);
				RND(ref e, a, ref b, c, d, F1, uintPtr[11], 15, 0);
				RND(ref d, e, ref a, b, c, F1, uintPtr[12], 6, 0);
				RND(ref c, d, ref e, a, b, F1, uintPtr[13], 7, 0);
				RND(ref b, c, ref d, e, a, F1, uintPtr[14], 9, 0);
				RND(ref a, b, ref c, d, e, F1, uintPtr[15], 8, 0);

				// round 2
				RND(ref e, a, ref b, c, d, F2, uintPtr[7], 7, 0x5A827999);
				RND(ref d, e, ref a, b, c, F2, uintPtr[4], 6, 0x5A827999);
				RND(ref c, d, ref e, a, b, F2, uintPtr[13], 8, 0x5A827999);
				RND(ref b, c, ref d, e, a, F2, uintPtr[1], 13, 0x5A827999);
				RND(ref a, b, ref c, d, e, F2, uintPtr[10], 11, 0x5A827999);
				RND(ref e, a, ref b, c, d, F2, uintPtr[6], 9, 0x5A827999);
				RND(ref d, e, ref a, b, c, F2, uintPtr[15], 7, 0x5A827999);
				RND(ref c, d, ref e, a, b, F2, uintPtr[3], 15, 0x5A827999);
				RND(ref b, c, ref d, e, a, F2, uintPtr[12], 7, 0x5A827999);
				RND(ref a, b, ref c, d, e, F2, uintPtr[0], 12, 0x5A827999);
				RND(ref e, a, ref b, c, d, F2, uintPtr[9], 15, 0x5A827999);
				RND(ref d, e, ref a, b, c, F2, uintPtr[5], 9, 0x5A827999);
				RND(ref c, d, ref e, a, b, F2, uintPtr[2], 11, 0x5A827999);
				RND(ref b, c, ref d, e, a, F2, uintPtr[14], 7, 0x5A827999);
				RND(ref a, b, ref c, d, e, F2, uintPtr[11], 13, 0x5A827999);
				RND(ref e, a, ref b, c, d, F2, uintPtr[8], 12, 0x5A827999);

				// round 3
				RND(ref d, e, ref a, b, c, F3, uintPtr[3], 11, 0x6ED9EBA1);
				RND(ref c, d, ref e, a, b, F3, uintPtr[10], 13, 0x6ED9EBA1);
				RND(ref b, c, ref d, e, a, F3, uintPtr[14], 6, 0x6ED9EBA1);
				RND(ref a, b, ref c, d, e, F3, uintPtr[4], 7, 0x6ED9EBA1);
				RND(ref e, a, ref b, c, d, F3, uintPtr[9], 14, 0x6ED9EBA1);
				RND(ref d, e, ref a, b, c, F3, uintPtr[15], 9, 0x6ED9EBA1);
				RND(ref c, d, ref e, a, b, F3, uintPtr[8], 13, 0x6ED9EBA1);
				RND(ref b, c, ref d, e, a, F3, uintPtr[1], 15, 0x6ED9EBA1);
				RND(ref a, b, ref c, d, e, F3, uintPtr[2], 14, 0x6ED9EBA1);
				RND(ref e, a, ref b, c, d, F3, uintPtr[7], 8, 0x6ED9EBA1);
				RND(ref d, e, ref a, b, c, F3, uintPtr[0], 13, 0x6ED9EBA1);
				RND(ref c, d, ref e, a, b, F3, uintPtr[6], 6, 0x6ED9EBA1);
				RND(ref b, c, ref d, e, a, F3, uintPtr[13], 5, 0x6ED9EBA1);
				RND(ref a, b, ref c, d, e, F3, uintPtr[11], 12, 0x6ED9EBA1);
				RND(ref e, a, ref b, c, d, F3, uintPtr[5], 7, 0x6ED9EBA1);
				RND(ref d, e, ref a, b, c, F3, uintPtr[12], 5, 0x6ED9EBA1);

				// round 4
				RND(ref c, d, ref e, a, b, F4, uintPtr[1], 11, 0x8F1BBCDC);
				RND(ref b, c, ref d, e, a, F4, uintPtr[9], 12, 0x8F1BBCDC);
				RND(ref a, b, ref c, d, e, F4, uintPtr[11], 14, 0x8F1BBCDC);
				RND(ref e, a, ref b, c, d, F4, uintPtr[10], 15, 0x8F1BBCDC);
				RND(ref d, e, ref a, b, c, F4, uintPtr[0], 14, 0x8F1BBCDC);
				RND(ref c, d, ref e, a, b, F4, uintPtr[8], 15, 0x8F1BBCDC);
				RND(ref b, c, ref d, e, a, F4, uintPtr[12], 9, 0x8F1BBCDC);
				RND(ref a, b, ref c, d, e, F4, uintPtr[4], 8, 0x8F1BBCDC);
				RND(ref e, a, ref b, c, d, F4, uintPtr[13], 9, 0x8F1BBCDC);
				RND(ref d, e, ref a, b, c, F4, uintPtr[3], 14, 0x8F1BBCDC);
				RND(ref c, d, ref e, a, b, F4, uintPtr[7], 5, 0x8F1BBCDC);
				RND(ref b, c, ref d, e, a, F4, uintPtr[15], 6, 0x8F1BBCDC);
				RND(ref a, b, ref c, d, e, F4, uintPtr[14], 8, 0x8F1BBCDC);
				RND(ref e, a, ref b, c, d, F4, uintPtr[5], 6, 0x8F1BBCDC);
				RND(ref d, e, ref a, b, c, F4, uintPtr[6], 5, 0x8F1BBCDC);
				RND(ref c, d, ref e, a, b, F4, uintPtr[2], 12, 0x8F1BBCDC);

				// round 5
				RND(ref b, c, ref d, e, a, F5, uintPtr[4], 9, 0xA953FD4E);
				RND(ref a, b, ref c, d, e, F5, uintPtr[0], 15, 0xA953FD4E);
				RND(ref e, a, ref b, c, d, F5, uintPtr[5], 5, 0xA953FD4E);
				RND(ref d, e, ref a, b, c, F5, uintPtr[9], 11, 0xA953FD4E);
				RND(ref c, d, ref e, a, b, F5, uintPtr[7], 6, 0xA953FD4E);
				RND(ref b, c, ref d, e, a, F5, uintPtr[12], 8, 0xA953FD4E);
				RND(ref a, b, ref c, d, e, F5, uintPtr[2], 13, 0xA953FD4E);
				RND(ref e, a, ref b, c, d, F5, uintPtr[10], 12, 0xA953FD4E);
				RND(ref d, e, ref a, b, c, F5, uintPtr[14], 5, 0xA953FD4E);
				RND(ref c, d, ref e, a, b, F5, uintPtr[1], 12, 0xA953FD4E);
				RND(ref b, c, ref d, e, a, F5, uintPtr[3], 13, 0xA953FD4E);
				RND(ref a, b, ref c, d, e, F5, uintPtr[8], 14, 0xA953FD4E);
				RND(ref e, a, ref b, c, d, F5, uintPtr[11], 11, 0xA953FD4E);
				RND(ref d, e, ref a, b, c, F5, uintPtr[6], 8, 0xA953FD4E);
				RND(ref c, d, ref e, a, b, F5, uintPtr[15], 5, 0xA953FD4E);
				RND(ref b, c, ref d, e, a, F5, uintPtr[13], 6, 0xA953FD4E);

				// right rounds
				uint aa = _h0;
				uint bb = _h1;
				uint cc = _h2;
				uint dd = _h3;
				uint ee = _h4;

				// round 1
				RND(ref aa, bb, ref cc, dd, ee, F5, uintPtr[5], 8, 0x50A28BE6);
				RND(ref ee, aa, ref bb, cc, dd, F5, uintPtr[14], 9, 0x50A28BE6);
				RND(ref dd, ee, ref aa, bb, cc, F5, uintPtr[7], 9, 0x50A28BE6);
				RND(ref cc, dd, ref ee, aa, bb, F5, uintPtr[0], 11, 0x50A28BE6);
				RND(ref bb, cc, ref dd, ee, aa, F5, uintPtr[9], 13, 0x50A28BE6);
				RND(ref aa, bb, ref cc, dd, ee, F5, uintPtr[2], 15, 0x50A28BE6);
				RND(ref ee, aa, ref bb, cc, dd, F5, uintPtr[11], 15, 0x50A28BE6);
				RND(ref dd, ee, ref aa, bb, cc, F5, uintPtr[4], 5, 0x50A28BE6);
				RND(ref cc, dd, ref ee, aa, bb, F5, uintPtr[13], 7, 0x50A28BE6);
				RND(ref bb, cc, ref dd, ee, aa, F5, uintPtr[6], 7, 0x50A28BE6);
				RND(ref aa, bb, ref cc, dd, ee, F5, uintPtr[15], 8, 0x50A28BE6);
				RND(ref ee, aa, ref bb, cc, dd, F5, uintPtr[8], 11, 0x50A28BE6);
				RND(ref dd, ee, ref aa, bb, cc, F5, uintPtr[1], 14, 0x50A28BE6);
				RND(ref cc, dd, ref ee, aa, bb, F5, uintPtr[10], 14, 0x50A28BE6);
				RND(ref bb, cc, ref dd, ee, aa, F5, uintPtr[3], 12, 0x50A28BE6);
				RND(ref aa, bb, ref cc, dd, ee, F5, uintPtr[12], 6, 0x50A28BE6);

				// round 2
				RND(ref ee, aa, ref bb, cc, dd, F4, uintPtr[6], 9, 0x5C4DD124);
				RND(ref dd, ee, ref aa, bb, cc, F4, uintPtr[11], 13, 0x5C4DD124);
				RND(ref cc, dd, ref ee, aa, bb, F4, uintPtr[3], 15, 0x5C4DD124);
				RND(ref bb, cc, ref dd, ee, aa, F4, uintPtr[7], 7, 0x5C4DD124);
				RND(ref aa, bb, ref cc, dd, ee, F4, uintPtr[0], 12, 0x5C4DD124);
				RND(ref ee, aa, ref bb, cc, dd, F4, uintPtr[13], 8, 0x5C4DD124);
				RND(ref dd, ee, ref aa, bb, cc, F4, uintPtr[5], 9, 0x5C4DD124);
				RND(ref cc, dd, ref ee, aa, bb, F4, uintPtr[10], 11, 0x5C4DD124);
				RND(ref bb, cc, ref dd, ee, aa, F4, uintPtr[14], 7, 0x5C4DD124);
				RND(ref aa, bb, ref cc, dd, ee, F4, uintPtr[15], 7, 0x5C4DD124);
				RND(ref ee, aa, ref bb, cc, dd, F4, uintPtr[8], 12, 0x5C4DD124);
				RND(ref dd, ee, ref aa, bb, cc, F4, uintPtr[12], 7, 0x5C4DD124);
				RND(ref cc, dd, ref ee, aa, bb, F4, uintPtr[4], 6, 0x5C4DD124);
				RND(ref bb, cc, ref dd, ee, aa, F4, uintPtr[9], 15, 0x5C4DD124);
				RND(ref aa, bb, ref cc, dd, ee, F4, uintPtr[1], 13, 0x5C4DD124);
				RND(ref ee, aa, ref bb, cc, dd, F4, uintPtr[2], 11, 0x5C4DD124);

				// round 3
				RND(ref dd, ee, ref aa, bb, cc, F3, uintPtr[15], 9, 0x6D703EF3);
				RND(ref cc, dd, ref ee, aa, bb, F3, uintPtr[5], 7, 0x6D703EF3);
				RND(ref bb, cc, ref dd, ee, aa, F3, uintPtr[1], 15, 0x6D703EF3);
				RND(ref aa, bb, ref cc, dd, ee, F3, uintPtr[3], 11, 0x6D703EF3);
				RND(ref ee, aa, ref bb, cc, dd, F3, uintPtr[7], 8, 0x6D703EF3);
				RND(ref dd, ee, ref aa, bb, cc, F3, uintPtr[14], 6, 0x6D703EF3);
				RND(ref cc, dd, ref ee, aa, bb, F3, uintPtr[6], 6, 0x6D703EF3);
				RND(ref bb, cc, ref dd, ee, aa, F3, uintPtr[9], 14, 0x6D703EF3);
				RND(ref aa, bb, ref cc, dd, ee, F3, uintPtr[11], 12, 0x6D703EF3);
				RND(ref ee, aa, ref bb, cc, dd, F3, uintPtr[8], 13, 0x6D703EF3);
				RND(ref dd, ee, ref aa, bb, cc, F3, uintPtr[12], 5, 0x6D703EF3);
				RND(ref cc, dd, ref ee, aa, bb, F3, uintPtr[2], 14, 0x6D703EF3);
				RND(ref bb, cc, ref dd, ee, aa, F3, uintPtr[10], 13, 0x6D703EF3);
				RND(ref aa, bb, ref cc, dd, ee, F3, uintPtr[0], 13, 0x6D703EF3);
				RND(ref ee, aa, ref bb, cc, dd, F3, uintPtr[4], 7, 0x6D703EF3);
				RND(ref dd, ee, ref aa, bb, cc, F3, uintPtr[13], 5, 0x6D703EF3);

				// round 4
				RND(ref cc, dd, ref ee, aa, bb, F2, uintPtr[8], 15, 0x7A6D76E9);
				RND(ref bb, cc, ref dd, ee, aa, F2, uintPtr[6], 5, 0x7A6D76E9);
				RND(ref aa, bb, ref cc, dd, ee, F2, uintPtr[4], 8, 0x7A6D76E9);
				RND(ref ee, aa, ref bb, cc, dd, F2, uintPtr[1], 11, 0x7A6D76E9);
				RND(ref dd, ee, ref aa, bb, cc, F2, uintPtr[3], 14, 0x7A6D76E9);
				RND(ref cc, dd, ref ee, aa, bb, F2, uintPtr[11], 14, 0x7A6D76E9);
				RND(ref bb, cc, ref dd, ee, aa, F2, uintPtr[15], 6, 0x7A6D76E9);
				RND(ref aa, bb, ref cc, dd, ee, F2, uintPtr[0], 14, 0x7A6D76E9);
				RND(ref ee, aa, ref bb, cc, dd, F2, uintPtr[5], 6, 0x7A6D76E9);
				RND(ref dd, ee, ref aa, bb, cc, F2, uintPtr[12], 9, 0x7A6D76E9);
				RND(ref cc, dd, ref ee, aa, bb, F2, uintPtr[2], 12, 0x7A6D76E9);
				RND(ref bb, cc, ref dd, ee, aa, F2, uintPtr[13], 9, 0x7A6D76E9);
				RND(ref aa, bb, ref cc, dd, ee, F2, uintPtr[9], 12, 0x7A6D76E9);
				RND(ref ee, aa, ref bb, cc, dd, F2, uintPtr[7], 5, 0x7A6D76E9);
				RND(ref dd, ee, ref aa, bb, cc, F2, uintPtr[10], 15, 0x7A6D76E9);
				RND(ref cc, dd, ref ee, aa, bb, F2, uintPtr[14], 8, 0x7A6D76E9);

				// round 5
				RND(ref bb, cc, ref dd, ee, aa, F1, uintPtr[12], 8, 0);
				RND(ref aa, bb, ref cc, dd, ee, F1, uintPtr[15], 5, 0);
				RND(ref ee, aa, ref bb, cc, dd, F1, uintPtr[10], 12, 0);
				RND(ref dd, ee, ref aa, bb, cc, F1, uintPtr[4], 9, 0);
				RND(ref cc, dd, ref ee, aa, bb, F1, uintPtr[1], 12, 0);
				RND(ref bb, cc, ref dd, ee, aa, F1, uintPtr[5], 5, 0);
				RND(ref aa, bb, ref cc, dd, ee, F1, uintPtr[8], 14, 0);
				RND(ref ee, aa, ref bb, cc, dd, F1, uintPtr[7], 6, 0);
				RND(ref dd, ee, ref aa, bb, cc, F1, uintPtr[6], 8, 0);
				RND(ref cc, dd, ref ee, aa, bb, F1, uintPtr[2], 13, 0);
				RND(ref bb, cc, ref dd, ee, aa, F1, uintPtr[13], 6, 0);
				RND(ref aa, bb, ref cc, dd, ee, F1, uintPtr[14], 5, 0);
				RND(ref ee, aa, ref bb, cc, dd, F1, uintPtr[0], 15, 0);
				RND(ref dd, ee, ref aa, bb, cc, F1, uintPtr[3], 13, 0);
				RND(ref cc, dd, ref ee, aa, bb, F1, uintPtr[9], 11, 0);
				RND(ref bb, cc, ref dd, ee, aa, F1, uintPtr[11], 11, 0);

				// combine the two parts
				dd += _h1 + c;
				_h1 = _h2 + d + ee;
				_h2 = _h3 + e + aa;
				_h3 = _h4 + a + bb;
				_h4 = _h0 + b + cc;
				_h0 = dd;
			}
		}

		private static uint F1(uint x, uint y, uint z)
		{
			return ((x) ^ (y) ^ (z));
		}

		private static uint ROL(uint x, uint n)
		{
			return (((x) << (int) (n)) | ((x) >> (int) (32 - (n))));
		}

		private static uint F2(uint x, uint y, uint z)
		{
			return (((x) & (y)) | (~(x) & (z)));
		}

		private static uint F3(uint x, uint y, uint z)
		{
			return (((x) | ~(y)) ^ (z));
		}

		private static uint F4(uint x, uint y, uint z)
		{
			return (((x) & (z)) | ((y) & ~(z)));
		}

		private static uint F5(uint x, uint y, uint z)
		{
			return ((x) ^ ((y) | ~(z)));
		}

		private static void RND(ref uint a, uint b, ref uint c, uint d, uint e, Func<uint, uint, uint, uint> f, uint x, uint s, uint k)
		{
			(a) += f(b, c, d) + (x) + (k);
			(a) = ROL(a, s) + (e);
			(c) = ROL(c, 10);
		}

		// Compute the digest of a single block.
	}
}