﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace E4GuiTool
{
	public partial class EditorForm : Form
	{
		private const string ValidHexChars = "0123456789ABCDEF";

		public EditorForm()
		{
			InitializeComponent();
		}

		private void OnSaveButton_Click(object sender, EventArgs e)
		{
			ulong suid = GetSuid();
			if (suid == 0)
			{
				MessageBox.Show(this, "Please make sure you have a valid 16 byte hex suid set.", "Invalid SUID", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
				return;
			}

			DialogResult r = saveFileDialog.ShowDialog(this);
			if (r != DialogResult.OK)
				return;

			string document = UpdateSumAndGenerateOutputDocument();
			byte[] documentBytes = Encoding.ASCII.GetBytes(document);
			using (var ms = new MemoryStream(documentBytes, false))
			using (FileStream fs = File.Open(saveFileDialog.FileName, FileMode.Create))
			{
				FTool.Encrypt(ms, fs, suid);
			}

			// Reload the document on save to reflext what was actually saved.
			editTextBox.Text = document;
		}

		private void OnOpenButton_Click(object sender, EventArgs e)
		{
			DialogResult r = openFileDialog.ShowDialog(this);
			if (r != DialogResult.OK)
				return;

			using (var ms = new MemoryStream())
			using (FileStream fs = File.OpenRead(openFileDialog.FileName))
			{
				try
				{
					ulong suid = FTool.Decrypt(fs, ms);
					SetSuid(suid);
					byte[] msBuffer = ms.GetBuffer();
					string text = Encoding.ASCII.GetString(msBuffer, 0, (int) ms.Position);
					editTextBox.Text = text;
				}
				catch (Exception ex)
				{
					MessageBox.Show(this, "Error: " + ex.Message, "Error opening file", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
		}

		private void OnEditTextBox_TextChanged(object sender, EventArgs e)
		{
			UpdateSumAndGenerateOutputDocument();
		}

		private void SetSuid(ulong suid)
		{
			string hexSuid = "0x" + suid.ToString("X16");
			suidTextBox.Text = hexSuid;
		}


		private ulong GetSuid()
		{
			string hexSuid = suidTextBox.Text;
			if (hexSuid.Length < 2)
				return 0;
			hexSuid = hexSuid.Substring(2);
			// Len check.
			if (hexSuid.Length != 16)
				return 0;
			// Invalid character check.
			if (hexSuid.Any(c => !ValidHexChars.Contains(c)))
				return 0;
			return Convert.ToUInt64(hexSuid, 16);
		}

		private string UpdateSumAndGenerateOutputDocument()
		{
			var sb = new StringBuilder();
			List<string> lines = editTextBox.Lines.ToList();

			// Removing any trailing empty lines.
			while (string.IsNullOrEmpty(lines[lines.Count - 1]))
				lines.RemoveAt(lines.Count - 1);

			// Remove any existing checksums.
			foreach (string line in lines.ToArray())
			{
				if (line.StartsWith("# CRC03"))
					continue;
				sb.AppendLine(line);
			}

			string document = sb.ToString();
			byte[] documentBytes = Encoding.ASCII.GetBytes(document);
			var sum = new Crmd160(3);
			sum.Update(documentBytes, documentBytes.Length);
			var digest = new byte[20];
			sum.Digest(digest);
			string digestText = string.Format("# CRC03 {0:x}{1:x}{2:x}{3:x}", digest[3] & 0xff, digest[2] & 0xff, digest[1] & 0xff, digest[0] & 0xff);
			sb.AppendLine(digestText);

			sumTextBox.Text = digestText;
			return sb.ToString();
		}

		private void CfcEditorForm_Shown(object sender, EventArgs e)
		{
			ShowAbout();
		}

		private void OnInfoButton_Click(object sender, EventArgs e)
		{
			ShowAbout();
		}


		private void ShowAbout()
		{
			using (var about = new AboutForm())
				about.ShowDialog();
		}
	}
}